/* LittleLights driver
 * kernel driver for LittleLights / BlinkenLEDs(Pro) / BLINKENmini
 * version 1.2.1 date 2005-07-11
 * Copyright (C) 2004-2005 ST <st@littlelights.de> & 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/ioport.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/param.h>
#include <linux/interrupt.h>
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
#  include <linux/pnp.h>
#endif

//number of elements in an array
#define count( array ) (sizeof( (array) ) / sizeof( (array)[0] ))

//version information for insmod
char version[] = UTS_RELEASE;

//startup parameters
int io = 0; //IO address
MODULE_PARM( io, "i" );
int irq = -1; //IRQ number (-1 for none)
MODULE_PARM( irq, "i" );
int hw = 0; //hardware type
MODULE_PARM( hw, "i" );
int gamma = 10; //gamma correction value
MODULE_PARM( gamma, "i" );

//global constants concerning the module
#define LlDrvDevNoMajor 0xB0 //the major device number for the littlelights-device-inode
#define LlDrvDevName "lldrv" //the name of this device (max. 31 chars allowed)
#define LlDrvDevId ((void *)(('l'<<24)+('l'<<16)+('d'<<8)+'r')) //the id of this device

//global types
#define LlDrvInternalPictureSizeX 18 //dimensions of an internal BM-picture
#define LlDrvInternalPictureSizeY 8
typedef struct tLlDrvInternalPicture //an internal BM-picture
{
  unsigned char Pixels[LlDrvInternalPictureSizeY][LlDrvInternalPictureSizeX];
} stLlDrvInternalPicture;

//global variables
int LlDrvIsOpen = 0; //true if the device is opened by a process
//IO address and IRQ number
unsigned short LlDrvIoAddrs[] = { 0x3BC, 0x378, 0x278 }; //valid io-addresses
unsigned short LlDrvIoAddr; //io-address
char LlDrvIrqNos[] = { -1, 7, 5 }; //valid irq-numbers (-1 for no irq)
char LlDrvIrqNo; //irq-number
//hardware type
#define LlDrvHwLl 0 //LittleLights, BlinkenLEDs
#define LlDrvHwBm 1 //BLINKENmini
unsigned char LlDrvHws[] = { LlDrvHwLl, LlDrvHwBm }; //valid hardware types
unsigned char LlDrvHw; //hardware type to use
//gamma correction (brightness-level to duty-cycle conversion)
#define LlDrvOutputSteps 15
struct LlDrvGamma{
  unsigned char Gamma;
  char PixelVals[LlDrvOutputSteps];
} LlDrvGammas[] =
{
  {  3, { 0x00, 0x00, 0x00, 0x02, 0x05, 0x09, 0x10, 0x19, 0x25, 0x35, 0x49, 0x62, 0x80, 0xA4, 0xCE } },
  {  5, { 0x01, 0x04, 0x09, 0x10, 0x19, 0x24, 0x31, 0x40, 0x51, 0x64, 0x79, 0x90, 0xA9, 0xC4, 0xE1 } },
  {  7, { 0x04, 0x0D, 0x17, 0x23, 0x30, 0x3F, 0x4E, 0x5F, 0x70, 0x82, 0x95, 0xA9, 0xBE, 0xD3, 0xE9 } },
  { 10, { 0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70, 0x80, 0x90, 0xA0, 0xB0, 0xC0, 0xD0, 0xE0, 0xF0 } },
  { 15, { 0x28, 0x40, 0x53, 0x65, 0x75, 0x85, 0x93, 0xA1, 0xAE, 0xBB, 0xC7, 0xD3, 0xDE, 0xEA, 0xF5 } },
  { 20, { 0x40, 0x5A, 0x6E, 0x80, 0x8F, 0x9C, 0xA9, 0xB5, 0xC0, 0xCA, 0xD4, 0xDD, 0xE6, 0xEF, 0xF7 } },
  { 30, { 0x65, 0x80, 0x92, 0xA1, 0xAD, 0xB8, 0xC2, 0xCB, 0xD3, 0xDA, 0xE1, 0xE8, 0xEE, 0xF4, 0xFA } }
};
unsigned char LlDrvGamma; //gamma valu in use
unsigned char * LlDrvOutputStepPixelVals; //pixel values to use

//the next step of the output-process to take when an interrupt occurs
int LlDrvOutputStep = 0;

//the current picture that is being displayed
struct tLlDrvInternalPicture LlDrvCurrentPicture;

//the next picture to display
//this picture exists two times
//one can be altered by the ioctl-syscall
//the other one can be read by the interrupt (which can happen at any time)
struct tLlDrvInternalPicture LlDrvNextPictures[2];
//the index of the picture that currently belongs to the interrupt
int LlDrvNextPictureInterruptIndex;

//the standard-picture: ll-logo.bmm
struct tLlDrvInternalPicture LlDrvStdPicture =
{
Pixels:
  {
    { 0x11, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF, 0xFF },
    { 0x33, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xDD },
    { 0x55, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xBB },
    { 0x77, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x99 },
    { 0x99, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x77 },
    { 0xBB, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x55 },
    { 0xDD, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x33 },
    { 0xFF, 0xFF, 0xEE, 0xDD, 0xCC, 0xBB, 0xAA, 0x99, 0x88, 0x88, 0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11, 0x11 }
   }
};

//buffer for frames written to the device
int LlDrvWriteFrame; //boolean flag if processing a frame
unsigned char LlDrvWriteBuf[64]; //buffer to read the frame header
int LlDrvWriteBufLen;
int LlDrvWriteHeight, LlDrvWriteWidth, LlDrvWriteChannels, LlDrvWriteMaxval; //parameters of frame being processed
int LlDrvWriteFillL, LlDrvWriteFillT, LlDrvWriteSkipL, LlDrvWriteSkipT;
int LlDrvWriteVal, LlDrvWriteC, LlDrvWriteX, LlDrvWriteY; //position in frame being processed

//output a frame (LittleLights style)
//turns on all pixels of *pPicture that have a value >= PixelValue
//if pPicture is NULL, all pixels are turned off
static void LlDrvFrameOutLl( struct tLlDrvInternalPicture * pPicture, unsigned char PixelValue )
{
  //local variables
  int x, y;
  unsigned char Val, Strobe;

  //loop through all 9 bit groups
  for( x = 1; x >= 0; x-- )
  for( y = LlDrvInternalPictureSizeY - 1; y >= 0; y-- )
  {
    //calculate value and strobe for this 9 pixels
    Val = 0x00;
    Strobe = 0x00;
    if( pPicture != NULL )
    {
      if( pPicture->Pixels[y][ 0 + x] >= PixelValue )
        Val |= 0x01;
      if( pPicture->Pixels[y][ 2 + x] >= PixelValue )
        Val |= 0x02;
      if( pPicture->Pixels[y][ 4 + x] >= PixelValue )
        Val |= 0x04;
      if( pPicture->Pixels[y][ 6 + x] >= PixelValue )
        Val |= 0x08;
      if( pPicture->Pixels[y][ 8 + x] >= PixelValue )
        Val |= 0x10;
      if( pPicture->Pixels[y][10 + x] >= PixelValue )
        Val |= 0x20;
      if( pPicture->Pixels[y][12 + x] >= PixelValue )
        Val |= 0x40;
      if( pPicture->Pixels[y][14 + x] >= PixelValue )
        Val |= 0x80;
      if( pPicture->Pixels[y][16 + x] >= PixelValue )
        Strobe |= 0x01;
    }

    //output the next 9 bits of data
    outb( Val, LlDrvIoAddr );
    outb( 0x13 & ~Strobe, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=!Strobe, -nAUTOLF=1, nINIT=0, -nSLCTIN=0

    //output a clock-pulse (on pin -nAUTOLF)
    outb( 0x11 & ~Strobe, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=!Strobe, -nAUTOLF=0, nINIT=0, -nSLCTIN=0
    outb( 0x13 & ~Strobe, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=!Strobe, -nAUTOLF=1, nINIT=0, -nSLCTIN=0
  } //for( y ... for( x ...

  //output a load-pulse (on pin nINIT)
  outb( 0x17, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=1, -nAUTOLF=1, nINIT=1, -nSLCTIN=0
  outb( 0x13, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=1, -nAUTOLF=1, nINIT=0, -nSLCTIN=0
}

//output a frame (BLINKENmini style)
//turns on all pixels of *pPicture that have a value >= PixelValue
//if pPicture is NULL, all pixels are turned off
static void LlDrvFrameOutBm( struct tLlDrvInternalPicture * pPicture, unsigned char PixelValue )
{
  //local variables
  int x, y;
  unsigned char Val;

  //loop through all columns
  //(backwards: leftmost pixel as last one)
  for( x = LlDrvInternalPictureSizeX - 1; x >= 0; x-- )
  {
    //calculate value for this column
    Val = 0;
    if( pPicture != NULL )
    {
      //loop through all pixels of this column
      //(backwards: topmost pixel into least significant bit)
      for( y = LlDrvInternalPictureSizeY - 1; y >= 0; y-- )
      {
        Val <<= 1;
        if( pPicture->Pixels[y][x] >= PixelValue )
          Val |= 1;
      }
    }

    //output the pixel-value for this column
    outb( Val, LlDrvIoAddr );

    //output a clock-pulse (on pin -nSTROBE)
    outb( 0x1A, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=0, -nAUTOLF=1, nINIT=0, -nSLCTIN=1
    outb( 0x1B, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=1, -nAUTOLF=1, nINIT=0, -nSLCTIN=1
  } //for( x ...

  //output a load-pulse (on pin -nAUTOLF)
  outb( 0x19, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=1, -nAUTOLF=0, nINIT=0, -nSLCTIN=1
  outb( 0x1B, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=1, -nAUTOLF=1, nINIT=0, -nSLCTIN=1
}

//output a frame
//turns on all pixels of *pPicture that have a value >= PixelValue
//if pPicture is NULL, all pixels are turned off
static void LlDrvFrameOut( struct tLlDrvInternalPicture * pPicture, unsigned char PixelValue )
{
  switch( LlDrvHw )
  {
    case LlDrvHwLl: //LittleLights, BlinkenLEDs
      LlDrvFrameOutLl( pPicture, PixelValue );
      break;
    case LlDrvHwBm: //BLINKENmini
      LlDrvFrameOutBm( pPicture, PixelValue );
      break;
  }
}

//the interrupt procedure
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
irqreturn_t LlDrvInt( int IrqNo, void * DevId, struct pt_regs * PtRegs )
#else
void LlDrvInt( int IrqNo, void * DevId, struct pt_regs * PtRegs )
#endif
{
  //check IRQ-number and device-id
  if( IrqNo != LlDrvIrqNo || DevId != LlDrvDevId )
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
    return IRQ_NONE;
#else
    return;
#endif

  //start of a new output-cycle
  if( LlDrvOutputStep == 0 )
    //copy next picture (that belongs to the interrupt) to current picture
    memcpy( &LlDrvCurrentPicture, &LlDrvNextPictures[LlDrvNextPictureInterruptIndex], sizeof( struct tLlDrvInternalPicture ) );

  //output the current frame of the current picture
  LlDrvFrameOut( &LlDrvCurrentPicture, LlDrvOutputStepPixelVals[LlDrvOutputStep] );

  //next output step
  LlDrvOutputStep++;
  if( LlDrvOutputStep >= LlDrvOutputSteps )
    LlDrvOutputStep = 0;

#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
  return IRQ_HANDLED;
#endif
}

//output system is notified of new picture
static void LlDrvNewFrame( void )
{
  //exit if using grayscales
  if( LlDrvIrqNo >= 0 )
    return;

  //output new picture as black/white
  LlDrvFrameOut( &LlDrvNextPictures[LlDrvNextPictureInterruptIndex], 0x80 );
}

//a process is opening the device
int LlDrvFileOpen( struct inode * Inode, struct file * File )
{
  //if output device is already open
  if( LlDrvIsOpen )
    //open failed
    return -EBUSY;

  //device is now open
  LlDrvIsOpen = 1;

  //no frame is being parsed
  LlDrvWriteFrame = 0;
  LlDrvWriteBufLen = 0;

  //empty next picture that belongs not to the interrupt
  memset( &LlDrvNextPictures[1 - LlDrvNextPictureInterruptIndex], 0, sizeof( struct tLlDrvInternalPicture ) );
  //make the new next picture the one that belongs to the interrupt
  //(because the assignment is an atomic instruction nothing can go wrong)
  //(the interrupt only reads this variable)
  //(this line can't be executed during execution of the interrupt)
  LlDrvNextPictureInterruptIndex = 1 - LlDrvNextPictureInterruptIndex;
  //notify output system of new picture
  LlDrvNewFrame( );

  //open succeeded
  return 0;  
}

//a process is closing the device
int LlDrvFileClose( struct inode * Inode, struct file * File )
{
  //if device is not open
  if( !LlDrvIsOpen )
    //close failed
    return -EBADF;

  //copy standard picture into next picture that belongs not to the interrupt
  memcpy( &LlDrvNextPictures[1 - LlDrvNextPictureInterruptIndex], &LlDrvStdPicture, sizeof( struct tLlDrvInternalPicture ) );
  //make the new next picture the one that belongs to the interrupt
  //(because the assignment is an atomic instruction nothing can go wrong)
  //(the interrupt only reads this variable)
  //(this line can't be executed during execution of the interrupt)
  LlDrvNextPictureInterruptIndex = 1 - LlDrvNextPictureInterruptIndex;
  //notify output system of new picture
  LlDrvNewFrame( );
 
  //device is now closed
  LlDrvIsOpen = 0;

  //no frame is being parsed
  LlDrvWriteFrame = 0;
  LlDrvWriteBufLen = 0;

  //close succeeded
  return 0;  
}

//a process is writing to the device
ssize_t LlDrvFileWrite( struct file * File, const char * p_buffer, size_t cnt, loff_t * offset_ptr )
{
  unsigned long magic;
  int i, x, y;
  unsigned char val;

  //if output device is not open
  if( !LlDrvIsOpen )
    //write failed
    return -EBADF;

  //test buffer
  if( verify_area( VERIFY_READ, p_buffer, cnt ) != 0 )
    //buffer not okay
    return -EFAULT;

  //processing loop
  for( i = 0; i < cnt; i++ )
  {

    //no frame yet
    if( ! LlDrvWriteFrame )
    {

      //put byte into buffer
      if( i < sizeof( LlDrvWriteBuf ) )
      {
        get_user( LlDrvWriteBuf[LlDrvWriteBufLen], ((unsigned char *)p_buffer + i) );
        LlDrvWriteBufLen++;
      }

      //entire header is in buffer
      if( LlDrvWriteBufLen == 12 )
      {

        //get magic
        magic = ntohl( *(unsigned long *)(LlDrvWriteBuf) );

        //get parameters
        switch( magic )
        {
          //DEADBEEF and FEEDBEEF frame
          case 0xDEADBEEF:
          case 0xFEEDBEEF:
            //get width and height
            LlDrvWriteWidth = ntohs( *(unsigned short *)(LlDrvWriteBuf + 8) );
            LlDrvWriteHeight = ntohs( *(unsigned short *)(LlDrvWriteBuf + 10) );
            //set rest of parameters
            LlDrvWriteChannels = 1;
            LlDrvWriteMaxval = magic == 0xFEEDBEEF ? 255 : 1;
            //parameters are ok
            if( LlDrvWriteHeight > 0 && LlDrvWriteHeight < 1000 &&
                LlDrvWriteWidth > 0 && LlDrvWriteWidth < 1000 )
              //empty buffer (and signal ok to next part of code)
              LlDrvWriteBufLen = 0;
            break;
          //MCU frame
          case 0x23542666:
            //get height and widtht
            LlDrvWriteHeight = ntohs( *(unsigned short *)(LlDrvWriteBuf + 4) );
            LlDrvWriteWidth = ntohs( *(unsigned short *)(LlDrvWriteBuf + 6) );
            //get channels and maxval
            LlDrvWriteChannels = ntohs( *(unsigned short *)(LlDrvWriteBuf + 8) );
            LlDrvWriteMaxval = ntohs( *(unsigned short *)(LlDrvWriteBuf + 10) );
            //parameters are ok
            if( LlDrvWriteHeight > 0 && LlDrvWriteHeight < 1000 &&
                LlDrvWriteWidth > 0 && LlDrvWriteWidth < 1000 &&
                LlDrvWriteChannels > 0 && LlDrvWriteChannels < 4 &&
                LlDrvWriteMaxval > 0 && LlDrvWriteMaxval < 256 )
              //empty buffer (and signal ok to next part of code)
              LlDrvWriteBufLen = 0;
            break;
        }

        //no valid packet header found
        if( LlDrvWriteBufLen > 0 )
        {
          //skip 1st character
          LlDrvWriteBufLen--;
          memmove( LlDrvWriteBuf, LlDrvWriteBuf + 1, LlDrvWriteBufLen );
        }
        //valid packet header found
        else
        {
          //calculate parameters
          LlDrvWriteFillL = LlDrvWriteWidth < LlDrvInternalPictureSizeX ? (LlDrvInternalPictureSizeX - LlDrvWriteWidth) / 2 : 0;
          LlDrvWriteFillT = LlDrvWriteHeight < LlDrvInternalPictureSizeY ? (LlDrvInternalPictureSizeY - LlDrvWriteHeight) / 2 : 0;
          LlDrvWriteSkipL = LlDrvWriteWidth > LlDrvInternalPictureSizeX ? (LlDrvWriteWidth - LlDrvInternalPictureSizeX) / 2 : 0;
          LlDrvWriteSkipT = LlDrvWriteHeight > LlDrvInternalPictureSizeY ? (LlDrvWriteHeight - LlDrvInternalPictureSizeY) / 2 : 0;
          //now processing frame
          LlDrvWriteFrame = 1;
          LlDrvWriteVal = 0;
          LlDrvWriteC = 0;
          LlDrvWriteX = 0;
          LlDrvWriteY = 0;
          //clear frame
          for( y = 0; y < LlDrvInternalPictureSizeY; y++ )
          for( x = 0; x < LlDrvInternalPictureSizeX; x++ )
            LlDrvNextPictures[1 - LlDrvNextPictureInterruptIndex].Pixels[y][x] = 0;
        }

      } //if( LlDrvWriteBufLen == 12 )

    } //if( ! LlDrvWriteFrame )

    //processing a frame
    else //if( ! LlDrvWriteFrame )
    {

      //get byte and convert it to rand 0x00 .. 0xFF
      get_user( val, ((unsigned char *)p_buffer + i) );
      LlDrvWriteVal += val > LlDrvWriteMaxval ? 0xFF : (0xFF * (unsigned int)val + LlDrvWriteMaxval / 2) / LlDrvWriteMaxval;

      //next channel
      LlDrvWriteC++;
      if( LlDrvWriteC >= LlDrvWriteChannels )
      {
        //get internal coordinates
        y = LlDrvWriteY - LlDrvWriteSkipT + LlDrvWriteFillT;
        x = LlDrvWriteX - LlDrvWriteSkipL + LlDrvWriteFillL;
        //if within visible range
        if( y >= 0 && y < LlDrvInternalPictureSizeY &&
            x >= 0 && x < LlDrvInternalPictureSizeX )
          //set pixel
          LlDrvNextPictures[1 - LlDrvNextPictureInterruptIndex].Pixels[y][x]
            = (unsigned char)((LlDrvWriteVal + LlDrvWriteChannels / 2) / LlDrvWriteChannels);

        LlDrvWriteVal = 0;
        LlDrvWriteC = 0;
        //next pixel
        LlDrvWriteX++;
        if( LlDrvWriteX >= LlDrvWriteWidth )
        {
          LlDrvWriteX = 0;
          //next line
          LlDrvWriteY++;
          if( LlDrvWriteY >= LlDrvWriteHeight )
          {
            //frame is completed - not processing a frame now
            LlDrvWriteFrame = 0;
            //make the new next picture the one that belongs to the interrupt
            //(because the assignment is an atomic instruction nothing can go wrong)
            //(the interrupt only reads this variable)
            //(this line can't be executed during execution of the interrupt)
            LlDrvNextPictureInterruptIndex = 1 - LlDrvNextPictureInterruptIndex;
            //notify output system of new picture
            LlDrvNewFrame( );
          } 
        }
      }

    } //if( ! LlDrvWriteFrame ) ... else

  } //for( i ...

  //return number of bytes written to this device
  return cnt;
}

//a process is reading from the output device
//reading from this device is always nonblocking - so when read returning 0 does not mean EOF
ssize_t LlDrvFileRead( struct file * File, char * BufferPtr, size_t BufferLen, loff_t * OffsetPtr )
{
  //if output device is not open
  if( !LlDrvIsOpen )
    //read failed
    return -EBADF;

  //test buffer
  if( verify_area( VERIFY_WRITE, BufferPtr, BufferLen ) != 0 )
    //buffer not okay
    return -EFAULT;

  //nothing written
  return 0;
}

//structure for supported operations
struct file_operations LlDrvFileOperations =
{
  .owner = THIS_MODULE,
  .read = LlDrvFileRead,
  .write = LlDrvFileWrite,
  .open = LlDrvFileOpen,
  .release = LlDrvFileClose,
  .llseek = NULL,
  .readdir = NULL,
  .poll = NULL,
  .ioctl = NULL,
  .mmap = NULL,
  .flush = NULL,
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
  .aio_read = NULL,
  .aio_write = NULL,
  .fsync = NULL,
  .aio_fsync = NULL,
  .fasync = NULL,
  .lock = NULL,
  .readv = NULL,
  .writev = NULL,
  .sendpage = NULL,
  .get_unmapped_area = NULL,
  .sendfile = NULL,
#endif
};

//PNP stuff for kernel 2.6
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
static const struct pnp_device_id LlDrvPnpTbl[] = {
  { .id = "PNP0400", .driver_data = 0 }, // standard printer port
  { .id = "PNP0401", .driver_data = 0 }, // ECP printer port
  { }
};
static int LlDrvPnpProbe( struct pnp_dev *dev, const struct pnp_device_id *id )
{
  if( pnp_port_valid( dev, 0 ) &&
      ! (pnp_port_flags( dev, 0 ) & IORESOURCE_DISABLED) &&
      pnp_port_start( dev , 0 ) == LlDrvIoAddr )
    return 0;
  else
    return -EINVAL;
}
static void LlDrvPnpRemove( struct pnp_dev *dev )
{
}
static struct pnp_driver LlDrvPnpDriver = {
  .name     = "lldrv",
  .id_table = LlDrvPnpTbl,
  .probe    = LlDrvPnpProbe,
  .remove   = LlDrvPnpRemove,
};
#endif

//module is being loaded
int init_module( void )
{
  //local variables
  int i;
  struct resource * ResPtr;

  //short message
  printk( KERN_INFO "lldrv: LittleLights driver\n" );
  printk( KERN_INFO "lldrv: kernel driver for LittleLights / BlinkenLEDs(Pro) / BLINKENmini\n" );
  printk( KERN_INFO "lldrv: version 1.2.1 date 2005-07-11\n" );
  printk( KERN_INFO "lldrv: Copyright (C) 2004-2005 ST <st@littlelights.de> & 1stein <1stein@schuermans.info>\n" );
  printk( KERN_INFO "lldrv: Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html\n" );
  printk( KERN_INFO "lldrv: trying to load (io=0x%X irq=%d hw=%d gamma=%d)...\n", io, irq, hw, gamma );

  //check startup parameters - io-address
  for( i = 0; i < count( LlDrvIoAddrs ); i++ )
    if( LlDrvIoAddrs[i] == (unsigned short)io )
      break;
  //invalid io-address
  if( i >= count( LlDrvIoAddrs ) )
  {
    //error message
    printk( KERN_ERR "lldrv: io=0x%X is not valid\n", io );
    printk( KERN_INFO "lldrv: valid settings for io are:" );
    for( i = 0; i < count( LlDrvIoAddrs ); i++ )
      printk( " 0x%X", LlDrvIoAddrs[i] );
    printk( "\n" );
    //init failed
    return -1;
  }
  //remember used io-address
  LlDrvIoAddr = LlDrvIoAddrs[i];

  //check startup parameters - irq-number
  for( i = 0; i < count( LlDrvIrqNos ); i++ )
    if( LlDrvIrqNos[i] == (char)irq )
      break;
  //invalid irq-number
  if( i >= count( LlDrvIrqNos ) )
  {
    //error message
    printk( KERN_ERR "lldrv: irq=%d is not valid\n", irq );
    printk( KERN_INFO "lldrv: valid settings for irq are:" );
    for( i = 0; i < count( LlDrvIrqNos ); i++ )
      printk( " %d", LlDrvIrqNos[i] );
    printk( " (-1 for none)\n" );
    //init failed
    return -1;
  }
  //remember used irq-number
  LlDrvIrqNo = LlDrvIrqNos[i];

  //check startup parameters - hardware type
  for( i = 0; i < count( LlDrvHws ); i++ )
    if( LlDrvHws[i] == (unsigned char)hw )
      break;
  //invalid irq-number
  if( i >= count( LlDrvHws ) )
  {
    //error message
    printk( KERN_ERR "lldrv: hw=%d is not valid\n", hw );
    printk( KERN_INFO "lldrv: valid settings for hw are:" );
    for( i = 0; i < count( LlDrvHws ); i++ )
      printk( " %d", LlDrvHws[i] );
    printk( "\n" );
    //init failed
    return -1;
  }
  //remember used hardware type
  LlDrvHw = LlDrvHws[i];
  
  //check startup parameters - gamma
  for( i = 0; i < count( LlDrvGammas ); i++ )
    if( LlDrvGammas[i].Gamma == (unsigned char)gamma )
      break;
  //invalid irq-number
  if( i >= count( LlDrvGammas ) )
  {
    //error message
    printk( KERN_ERR "lldrv: gamma=%d is not valid\n", gamma );
    printk( KERN_INFO "lldrv: valid settings for gamma are:" );
    for( i = 0; i < count( LlDrvGammas ); i++ )
      printk( " %d", LlDrvGammas[i].Gamma );
    printk( "\n" );
    //init failed
    return -1;
  }
  //remember gamma value
  LlDrvGamma = gamma;
  //save greyscale levels
  LlDrvOutputStepPixelVals = LlDrvGammas[i].PixelVals;

  //output and input device is not opend by a process yet
  LlDrvIsOpen = 0;

#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
  //register pnp driver
  pnp_register_driver( &LlDrvPnpDriver );
#endif

  //register device
  i = register_chrdev( LlDrvDevNoMajor, LlDrvDevName, &LlDrvFileOperations );
  //an error occured
  if( i < 0 )
  {
    //error-message
    printk( KERN_ERR "lldrv: register_chrdev failed with %d\n", i );
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
    //unregister pnp driver
    pnp_unregister_driver( &LlDrvPnpDriver );
#endif    
    //init failed
    return -1;
  }

  //register io-ports
  ResPtr = request_region( LlDrvIoAddr, 3, LlDrvDevName );
  //not free
  if( ResPtr == NULL )
  {
    //error-message
    printk( KERN_ERR "lldrv: request_region failed: io-ports not free\n" );
    //unregister device
    unregister_chrdev( LlDrvDevNoMajor, LlDrvDevName );
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
    //unregister pnp driver
    pnp_unregister_driver( &LlDrvPnpDriver );
#endif    
    //init failed
    return -1;
  }

  //setup hardware
  outb( 0x00, LlDrvIoAddr ); //set data-bits to 0
  outb( 0x13, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=1, -nAUTOLF=1, nINIT=0, -nSLCTIN=0

  //set the next picture to the standard picture
  //(nothing can go wrong, because interrupt is not active)
  memcpy( &LlDrvNextPictures[0], &LlDrvStdPicture, sizeof( struct tLlDrvInternalPicture ) ); //set picture 0
  LlDrvNextPictureInterruptIndex = 0; //picture 0 belongs to interrupt
  //start outputting the next picture
  //(nothing can go wrong, because interrupt is not active)
  LlDrvOutputStep = 0;
  //notify output system of new picture
  LlDrvNewFrame( );

  //if using grayscales
  if( LlDrvIrqNo >= 0 )
  {
    //register hardware-interrupt (not shareable)
    i = request_irq( LlDrvIrqNo, LlDrvInt, 0, LlDrvDevName, LlDrvDevId );
    //an error occured
    if( i != 0 )
    {
      //error-message
      printk( KERN_ERR "lldrv: request_irq failed with %d\n", i );
      //reset hardware
      outb( 0x00, LlDrvIoAddr ); //set data-bits to 0
      outb( 0x03, LlDrvIoAddr + 2 ); //bidirectional: off, irq: off, -nSTROBE=1, -nAUTOLF=1, nINIT=0, -nSLCTIN=0
      //unregister io-ports
      release_region( LlDrvIoAddr, 3 );
      //unregister device
      unregister_chrdev( LlDrvDevNoMajor, LlDrvDevName );
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
      //unregister pnp driver
      pnp_unregister_driver( &LlDrvPnpDriver );
#endif    
      //init failed
      return -1;
    }
  }

  //init succeeded
  printk( KERN_INFO "lldrv: loaded (io=0x%X irq=%d hw=%d gamma=%d)...\n", LlDrvIoAddr, LlDrvIrqNo, LlDrvHw, LlDrvGamma );
  return 0;
}

//module is being unloaded
void cleanup_module( void )
{
  //if using grayscales
  if( LlDrvIrqNo >= 0 )
    //unregister hardware-interrupt
    free_irq( LlDrvIrqNo, LlDrvDevId );

  //empty next picture that belongs not to the interrupt
  memset( &LlDrvNextPictures[1 - LlDrvNextPictureInterruptIndex], 0, sizeof( struct tLlDrvInternalPicture ) );
  //make the new next picture the one that belongs to the interrupt
  //(because the assignment is an atomic instruction nothing can go wrong)
  //(the interrupt only reads this variable)
  //(this line can't be executed during execution of the interrupt)
  LlDrvNextPictureInterruptIndex = 1 - LlDrvNextPictureInterruptIndex;
  //notify output system of new picture
  LlDrvNewFrame( );

  //output empty frame
  LlDrvFrameOut( NULL, 0 );

  //reset hardware
  outb( 0x00, LlDrvIoAddr ); //set data-bits to 0
  outb( 0x03, LlDrvIoAddr + 2 ); //bidirectional: off, irq: off, -nSTROBE=1, -nAUTOLF=1, nINIT=0, -nSLCTIN=0

  //unregister io-ports
  release_region( LlDrvIoAddr, 3 );

  //unregister device
  unregister_chrdev( LlDrvDevNoMajor, LlDrvDevName );

#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
  //unregister pnp driver
  pnp_unregister_driver( &LlDrvPnpDriver );
#endif    

  //done
  printk( KERN_INFO "lldrv: unloaded...\n" );
}

